#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Nov 11 15:14:15 2021

@author: aditimemani
"""
#from val_derv import val_derv
#from forward_mode import combine_vector_inputs, forward_mode
#from optimizers import Optimizer
